#! /usr/bin/env python
"""This script is similar to Maxine's "max" script. It enables us to quickly execute
the MBS interpreter on top of Maxine or the system default JVM.
"""
import os, sys, time
import fnmatch

from optparse import OptionParser

from pprint import pprint
from itertools import product
from measure import measure             ## is from the computer language shootout game...
from tempfile import mkstemp
from collections import defaultdict
import shlex
import datetime
import pickle
from math import *
import shutil
import csv
import numpy
import glob
# pip install scipy is currently broken on Lion
#import scipy.stats

def get_env():
    """This function returns a dictionary that contains the actual values for environment
    variables. This is platform specific (since we are relying on the platform identifier
    in Maxine's build directory. Currently, Linux [verified on Ubuntu 10.04+] and MacOS X
    are supported.
    NOTE: This function is not supposed to be called directly, rather use one of the pre-
    defined access functions that bind the environment and the key as a default argument,
    e.g.: get_maxine_executable().
    """
    currentDir = os.path.abspath('.')
    jythonDir = os.path.abspath('..')
    jythonProjectPath = jythonDir

    platform= sys.platform ## returns 'darwin' on macosx, 'linux2' on ubuntu 10.04
    if platform == 'linux2':
        platform= 'linux'

    #maxineExe = maxineDir + "/com.oracle.max.vm.native/generated/%s/maxvm -Xss512m -Xms4g -Xmx4g" % platform
    #maxineExe = maxineDir + "/com.oracle.max.vm.native/generated/%s/maxvm " % platform
    #maxineInspector = maxineDir + "/mxtool/mx inspect"
    #maxineInspector = maxineDir + "/mxtool/mx -d inspect"
    #maxineInspector = maxineDir + "/mx inspect"

    jythonScriptsPath = currentDir + "/benchmarks"

    classpath = ()

    # Jython stuff
    #jythonDir = os.path.abspath(currentDir, '..')
    jythonDist = os.path.join(jythonDir, 'dist')
    jythonDistJavaLib = os.path.join(jythonDist, 'javalib')

    if os.path.exists(jythonDistJavaLib):
        os.chdir(jythonDistJavaLib)
        classpath = [os.path.join(jythonDistJavaLib, cp) for cp in glob.glob('*.jar')]
        os.chdir(jythonDist)
        jythonJarFile = os.path.join(jythonDist, glob.glob('jython-dev.jar')[0])
        classpath.append(jythonJarFile)
        classpath = tuple(classpath)
        os.chdir(currentDir)

    return {
#        'executable' : maxineExe,
#        'maxinspect' : maxineInspector,
        'prj'        : jythonProjectPath,
        'scripts'    : jythonScriptsPath,
        'cp'         : classpath,
        'current'    : currentDir,
#        'maxine'     : maxineDir,
        }


#def get_maxine_home(env=get_env(), key='maxine'):
#    return env[key]
#
#def get_maxine_executable(env=get_env(), key='executable'):
#    return env[key]
#
#def get_maxine_inspector(env=get_env(), key='maxinspect'):
#    return env[key]

def get_maxine_prj(env=get_env(), key='prj'):
    return env[key]

def get_maxine_scripts(env=get_env(), key='scripts'):
    return env[key]

def get_maxine_classpath(env=get_env(), key='cp'):
    return ':'.join(env[key])

def get_cwd(env=get_env(), key='current'):
    return env[key]

def check_config():
    """This function verifies that some paths exist in the current configuration. If
    it fails, it reports an error to the caller and exits.
    """
    rules= [
        (get_maxine_prj(), "Invalid project path"),
        (get_maxine_scripts(), "Invalid scripts path"),
        ]
    for (p, errmsg) in rules:
        if not os.path.exists( p ):
            exit( "%s: %s" % (errmsg, p) )


def parse_cmd_line_args():
    """This defines the command line arguments supported by our tool.
    """
    p = OptionParser()

    p.add_option("-t", "--test",    action="store",      dest="test",    default="",    help="test {" + "}")
    p.add_option("-o", "--host",    action="store",      dest="host",    default="",    help="options to pass to the Host VM")
    p.add_option("-O", "--client",  action="store",      dest="client",  default="",    help="options to pass to the Client VM")
    p.add_option("-x", "--execute", action="store",      dest="execute", default="",    help="executes a specified script: {" + ", ".join(os.listdir(get_maxine_scripts())) + "}")
    p.add_option("-c", "--compile", action="store_true", dest="compile", default=False, help="compile all scripts")
    p.add_option("-r", "--run",     action="store_true", dest="run",     default=False, help="run all scripts")
    p.add_option("-m", "--max",     action="store_true", dest="max",     default=False, help="run on top of maxine")
    p.add_option("-i", "--ins",     action="store_true", dest="ins",     default=False, help="run on top of maxine inspector")
    p.add_option("-u", "--unmarshall",
                                    action="store_true", dest="unmarshall", default=False, help="Unmarshall a .pyc file and run in on bytecode interpreter")
    # p.add_option("-c", "--check", action="store_true", dest="check",   default=False, help="commit check")
    p.add_option("-b", "--build",   action="store_true", dest="build",   default=False, help="clean and build")
    p.add_option("-z", "--bench",   action="store",      dest="bench",   type='int',    default=0, help="run benchmarks and print results")

    return p.parse_args()


def get_exec(cmd_class="org.python.util.jython", action_args=[]):
    """This function returns a closure that holds partially bound arguments for executing
    programs using the "os.system" invocation function. The closure is used so that we can
    late bind arguments depending on the passed command line options. Furthermore, the when
    we execute the closure it will keep track of the exit codes and populate the corresponding
    "errors" list.
    """
    def exec_closure(args=[], host_vm_opts='', client_vm_opts='', use_maxine=False, use_inspector=False, errors=[]):
        # appending default host vm opts
        #host_vm_opts += ' -C1X:+ExceptionHandlingElimination -C1X:EHEFilter=Lorg/python/core'
        executable= 'java'

        if use_inspector:
            executable = get_maxine_inspector()
            host_vm_opts = ''
            os.chdir(get_maxine_home())
        elif use_maxine:
            executable = get_maxine_executable()

        if opts.unmarshall:
            # action_args += 'c'
            # special trick to keep sys.argv[1] in the unmarshalled scripts to
            # be the real argument we want to pass to the benchmark not the
            # bootstrapping /  unmarshalling wrapper script
            # also that .py is patched to be .pyc in the script name
            rearranged_action_args = ''.join(action_args)
            rearranged_action_args_list = rearranged_action_args.split(' ')
            rearranged_action_args_list.reverse()
            arguments = ' '.join(args + rearranged_action_args_list) + 'c'
            unmarshalScript = os.path.join(get_maxine_scripts(), 'unmarshall.py')
            arguments = unmarshalScript + ' ' + arguments
        else:
            arguments= ' '.join(action_args + args)

        command= "%s %s -cp %s %s %s %s > /dev/null" % (executable,
                                         host_vm_opts,
                                         get_maxine_classpath(),
                                         cmd_class,
                                         client_vm_opts,
                                         arguments)
        start= time.time()
        exit_code= os.system( command )
        elapsed= time.time() - start
        if exit_code:
            errors.append( (executable, host_vm_opts, client_vm_opts, cmd_class, arguments) )

	print '%5.3f' % elapsed

    return exec_closure


def do_build(prj=get_maxine_prj()):
    """Uses a simple list of rule-tuples to perform the actions to verify that the
    setup is actually working. Takes care of properly re-building the ANTLR generated
    resources, too.
    """
    rules= [
        ('Cleaning Maxine ...', [ './max clean' ]),
        ('Building Maxine ...', [ './max build' ]),
        ('Creating Maxine Image ...', [ './max image' ]),
        ]

    for (caption, actions) in rules:
        print caption
        sys.stdout.flush()
        for a in actions:
            os.system( a )
        print "Done"
    return []                           ## satisfy iterative interface

def enum_scripts(filter_fn=lambda x: True, script_path=get_maxine_scripts()):
    """This function is used to enumerate the scripts found in the script path. At the
    same time you can supply a custom "filter_fn" function to filter out unwanted entries
    from the input directory.
    NOTE: If this function finds a file called ".mbsignore" in the directory specified by
    the named "script_path" argument, it will exclude all files listed in there.
    """
    files_to_run= "%s/.run" % (script_path)

    if not os.path.exists(files_to_run):
        exit(files_to_run +  " does not exist")

    with open(files_to_run) as input:
        for s in input:
            if filter_fn(s):
                script = os.path.join(script_path, s.strip())
                yield script

def compile_gn():
    """This generator function binds the "-execute=false" argument to the execution
    closure.
    """
    arguments= ['-execute=false']
    for s in enum_scripts():
        yield get_exec(action_args=arguments + [s])

def run_gn(script_path=os.path.abspath(get_maxine_scripts())):
    for s in enum_scripts(filter_fn= lambda x: not x.startswith('#') ):
        if opts.unmarshall:
            print "Unmarshalling ... %s" % (os.path.basename(s))
        else:
            print "Running ... %s" % (os.path.basename(s))

        yield get_exec(action_args=[s])


def get_output_filename():
    base_dir= get_cwd()
    base_ext= "-mbs-bmark-results.txt"
    now= datetime.datetime.now();
    identifier= base_dir
    identifier+= "%4d%02d%02d" % (now.year, now.month, now.day)
    identifier+= "-"
    identifier+= "%02d%02d" % (now.hour, now.minute)
    identifier+= base_ext
    return identifier

def get_current_time_string():
    now = datetime.datetime.now()
    time_string = "%4d%02d%02d" % (now.year, now.month, now.day)
    time_string += "-"
    time_string += "%02d%02d" % (now.hour, now.minute)
    return time_string

def expand_file_information(tuples):
    """This function takes in a list of tuples as generated by "run_bmarks_gn" and
    expands the information stored within the temporary files generated by "run_bmarks_gn".
    The protocol is that the last line contains the running time of the benchmark as calculated
    in the benchmark (this is rather fuzzy though, as it is still in the interpreter and should
    actually be measured on the virtual machine level; probably a more accurate measurement
    technique would be to measure whole process execution as stored in the measure record "r" and
    subtract the VM start-up time.). All other information present in the file can be used for
    verification purposes, i.e., the results can later be compared to the outputs of other
    interpreters.
    (NOTE: printing to stdout is going to make most of the benchmark measurements useless)

    Finally, we generate a new list of tuples, with the information conveniently expanded
    for use by other functions.
    """
    results= []
    for (native, host, s, path, r) in tuples:
        with open(path) as f:
            lines= f.readlines()
            try:
                # if last line is time, the second last line is also ignored for the timer label
                prog_output, prog_runtime= lines[:-2], float(lines[-1])
                results.append( (native, host, s, prog_runtime, prog_output, path, r) )
            except:
#                print "Could not extract the timestamp information as expected, please fix benchmark!"
#                pprint( (native, host, s, path) )
#                print "Output is:"
#                print lines
                # bypass time extraction and carry on
                # if last line is not time, all lines are taken
                prog_output, prog_runtime= lines, float(0)
                results.append( (native, host, s, prog_runtime, prog_output, path, r) )

    return results

def build_tree(tuples):
    """This function builds an intermediate tree out of two dimensions of dictionaries to enable
    sub-sequent processing of data grouped by benchmarks and host virtual machines.
    It is an *internal* function and used to generate the tree argument used in  "flatten_tree",
    "rows_by_bmark", "verify_bmark_outputs", and "print_results_by_time".
    """
    tree= {}
    for (native, host, s, run_time, output, path, r) in tuples:
        if s not in tree:
            tree[s]= {}
        if host not in tree[s]:
            tree[s][host]= []
        tree[s][host].append( (run_time, output, native) )
    return tree

def rows_by_bmark(tree, selector=lambda x: x):
    """This generator uses a selector function to select data elements from the tuple and yield
    this selection together with the host and benchmark names. This simplifies sub-sequent
    processing in the "flatten_tree" function.
    """
    for bmark, bmark_data in tree.iteritems():
        outputs= []
        for host, host_data in bmark_data.iteritems():
            for elem in host_data:
                yield (bmark, host, selector(elem))

def flatten_tree(tree, selector):
    """This function flattens the tree from two dimensions into one, where the values for each key
    is a list that contains all of the elements found.
    """
    table= {}
    for (bmark, host, elem) in rows_by_bmark(tree, selector):
        if bmark not in table:
            table[bmark]= []
        table[bmark].append( (host, elem) )
    return table


def verify_bmark_outputs(tree):
    outputs= flatten_tree( tree, lambda (t,o,n): o)

    for bmark, data in outputs.iteritems():
        for ((lH, lO), (rH, rO)) in zip(data, data[opts.bench:]):
            if lO != rO:
                print "verify_bmark_outputs: %s: left: %s != right %s" % (bmark, lH, rH)
                print "BEGIN left\n%s\nEND left" % lO
                print "BEGIN right\n%s\nEND right" % rO

def escape_r(v):
    if v == None:
        return "NA"
    try:
        return str(float(v))
    except:
        return "\"" + str(v) + "\""

def write_frame(frame):
    """
    archive the old frame.r file
    """
    frame_file_name = "frame.r"
    oldframepath = os.path.join(os.getcwd(), frame_file_name)
    ctime = os.path.getctime(oldframepath)
    arcframename = ("frame-%s.r" % (time.ctime(ctime))).replace(' ', '_')
    arcframepath = os.path.join(os.getcwd(), arcframename)
    shutil.copyfile(oldframepath, arcframepath)


    with open("frame.r", 'w') as output:
        for k, v in frame.items():
            output.write("%s <- c(%s)\n" % (k, (','.join([escape_r(x) for x in v]))))
        output.write("data <- data.frame(%s)\n" % ','.join(frame.keys()))

def chart_results_by_time(tree):
    results= flatten_tree( tree, lambda (t,o,n): (t, n))
    frame = defaultdict(list)
    for bmark, data in results.iteritems():
        #print "\t%s:" % (bmark)
        for (host, (time, native)) in data:
            #print (host, (time, native))
            frame["Names"].append(os.path.split(bmark)[1])
            frame["Configuration"].append("%s-%s" % (native[0], host[0]))
            baseline = float(data[0][1][0])
            if baseline > 0 and float(time) > 0:
                frame["Time"].append(str(baseline / float(time)))
            else:
                frame["Time"].append("0")
    #pprint(frame)
    write_frame(frame)
    return


def compact_results(tree):
    results = flatten_tree(tree, lambda(t, o, n): (t, n))
#    print results
    compacted_results = []
    for bmark, data in results.iteritems():
        bmark = os.path.split(bmark)[1]
        result = []
        result.append(bmark)

        #print "bmark: " + bmark


        times = []
        index = 1
        for (host ,(time, native)) in data:
            #print time
            if index < opts.bench:
                times.append(time)
                index += 1
            else:
                times.append(time)
                result.append(times)
                times = []
                index = 1



        # calculating standard deviation and variance
        # now result looks like this
        # ['scriptName', [list of time for host0], [list of time for host0]]
        list_iter = iter(result)
        first_item_is_script_name = list_iter.next()
        averaged_result = [first_item_is_script_name]

        for time_list in list_iter:
            mean = numpy.mean(time_list)
            std = numpy.std(time_list) / mean
            variance = numpy.var(time_list) / mean**2
            averaged_result.append(mean)
            averaged_result.append(std)
            averaged_result.append(variance)

#        compacted_results.append(tuple(result))
        compacted_results.append(averaged_result)

    compacted_results.sort()
    table_header = []
    table_header.append('benchmark')

    hosts = []
    for (host ,(time, native)) in data:
        (host_name, ___, ___ ) = host
        host_native_name = host_name + '/' + native[0]
        if host_native_name not in hosts:
            hosts.append(host_native_name)

    for each_host in hosts:
        table_header.append(each_host)
        table_header.append('std')
        table_header.append('var')

    #table_header.append(data[0][1][1][0] + ':' + str(data[0][0][0]))
    #table_header.append(data[1][1][1][0] + ':' + str(data[1][0][0]))

    compacted_results.insert(0, tuple(table_header))
    return compacted_results


#Chen's Function to Print Out Raw Data
def print_raw(tree):
    results = flatten_tree(tree, lambda(t, o, n): (t, n))

    raw_results = []
    for bmark, data in results.iteritems():
        bmark = os.path.split(bmark)[1]
        result = []
        result.append(bmark)

        times = []
        #index = 1
        for (host ,(time, native)) in data:
            times.append(time)

        result.append(times)

        raw_results.append(result)


    table_header = []

    hosts = []
    for (host ,(time, native)) in data:
        (host_name, ___, ___ ) = host
        host_native_name = host_name + '/' + native[0]
        if host_native_name not in hosts:
            hosts.append(host_native_name)

    for each_host in hosts:
        table_header.append('benchmark ' + each_host)


    #print "-------------------------------------------------------------------------------------------------------------------------------"
    time_banner = "[" + str(opts.bench) + "X] " + str(datetime.datetime.now())
    #print "%127s" % time_banner
    #print "-------------------------------------------------------------------------------------------------------------------------------"

    index = 0
    host_output = []
    header = []
    header.append("type")
    header.append("benchmark")
    header.append("time")
    for each_host in hosts:
        title = 'benchmark [' + each_host + ']'
        #print "%29s" % (title)
        #print "-------------------------------------------------------------------------------------------------------------------------------"
        for result in raw_results:
            (bmark, times) = result

            format = ''
            format += '%29s\t'
            for i in range(1, opts.bench+1):
                format += '%15.2f'

            output = []
            output.append(bmark)

            for i in range(index, index + opts.bench):
            	raw_data = []
            	raw_data.append(each_host)
            	raw_data.append(bmark)
            	raw_data.append(times[i])
            	host_output.append(raw_data)
            	output.append(times[i])

            #print format % tuple(output)
        index = index + opts.bench



    #print "-------------------------------------------------------------------------------------------------------------------------------"
    result_file_base_name = get_current_time_string() + '-' + 'raw_result.csv'
    with open(result_file_base_name, "w") as output:
    	writer = csv.writer(output)
    	for result in host_output:
    		writer.writerow(result)

    last_raw_file_name = 'raw_result.csv'
    with open(last_raw_file_name, "w") as output:
    	writer = csv.writer(output)
    	for result in host_output:
    		writer.writerow(result)
#end print_raw




def print_compacted_results(results):
    print "-------------------------------------------------------------------------------------------------------------------------------"
    time_banner = "[" + str(opts.bench) + "X] " + str(datetime.datetime.now())
    print "%127s" % time_banner
    print "-------------------------------------------------------------------------------------------------------------------------------"
    (label, vm1, std1, var1, vm2, std2, var2) = results.pop(0)
#    swap= False
#
#    if [vm1, vm2] != ordering:
#    	vm1, std1, var1, vm2, std2, var2= vm2, std2, var2, vm1, std1, var1
#    	swap= True

    print "%29s\t%15s%10s%10s\t%15s%10s%10s\t%15s" % (label, '[' + vm1 + ']', std1, var1, '[' + vm2 + ']', std2, var2, 'speedup:2/1')
    print "-------------------------------------------------------------------------------------------------------------------------------"
    time1s, time2s = [], []
    speedups = []
    for result in results:
        (bmark, time1, std1, var1, time2, std2, var2) = result
#        if swap:
#            time1, time2= time2, time1
        speedup = time1/time2
        print "%29s\t%15.2f%10.2f%10.2f\t%15.2f%10.2f%10.2f\t%15f" % (bmark, time1, std1, var1, time2, std2, var2, speedup)
        time1s.append(time1)
        time2s.append(time2)
        speedups.append(speedup)

    print "-------------------------------------------------------------------------------------------------------------------------------"
    print "%29s\t%15.2f\t%39.2f\t%39f" % ('arithmetic mean',
                            float(sum(time1s)/len(time1s)),
                            float(sum(time2s)/len(time2s)),
                            float(sum(speedups)/len(speedups)))

    gmean_time1 = exp(sum([log(time1) for time1 in time1s])/len(time1s))
    gmean_time2 = exp(sum([log(time2) for time2 in time2s])/len(time2s))
    gmean_speedup = exp(sum([log(speedup) for speedup in speedups])/len(speedups))
    print "%29s\t%15.2f\t%39.2f\t%39f" % ('geometric mean', gmean_time1, gmean_time2, gmean_speedup)
    print "-------------------------------------------------------------------------------------------------------------------------------"
    # put header back to results
    results.insert(0, (label, vm1, vm2))
#    print results

def write_compacted_result_to_csv(results):
    result_file_base_name = 'result.csv'
    if os.path.isfile(result_file_base_name):
    	os.rename(result_file_base_name, get_current_time_string() + '-' + result_file_base_name)

    with open("result.csv", "w") as output:
        writer = csv.writer(output)
        for result in results:
            writer.writerow(result)

def print_results_by_time(tree):
    results= flatten_tree( tree, lambda (t,o,n): (t, n))

    print "############################## BENCHMARK results:"
    for bmark, data in results.iteritems():
        print "\t%s:" % (bmark)
        for (host, (time, native)) in sorted(data, key=lambda (h, (t, n)): -t):
            print "\t\t% 10.6f: %s (%s)" % (time, host, native)


def remove_temp_files(tuples):
    for (_, _, _, _, _, path, _) in tuples:
        if os.path.exists(path):
            os.remove(path)

def run_bmarks_gn(script_path=os.path.abspath(get_maxine_scripts())):
    native_vm_runtimes= [
          ("hotspot", "java")
          , ("hotspot", "java")
        ]

    host_vms= [
          ("bi", "org.python.util.jython " + os.path.join(get_maxine_scripts(), 'unmarshall.py'), "unmarshal")
          , ("bi", "org.python.util.jython " + os.path.join(get_maxine_scripts(), 'unmarshall.py'), "unmarshal")
        ]

    results= []
    try:
        index= 0
        with open('/dev/null', 'w') as throwaway:
            for (native, host) in zip(native_vm_runtimes, host_vms):
                print "Benchmarking %s on %s" % (host, native)
                for s in enum_scripts(filter_fn= lambda x: not x.startswith('#') ):
                    job = s
                    if host[2] == "unmarshal":
                        # reverse the job string when unmarshalling
                        script_list = s.split(' ')
                        script_list.reverse()
                        job = ' '.join(script_list) + 'c'
                        print "\tunmarshalling %35s" % (job.split(' ')[0] + ' ' + job.split('/')[-1]),
                    else:
                        print "\trunning %35s" % (job.split('/')[-1]),

                    cmd_line= "%s -cp %s %s %s" % (native[1],
                                                   get_maxine_classpath(),
                                                   host[1],
                                                   job)
                    cmd= shlex.split(cmd_line)
                    print " -- ",
                    for iteration in range(0, opts.bench):
                        print "%d," % iteration,
                        sys.stdout.flush()
                        retry= True
                        while retry:
                            (output, path)= mkstemp(prefix="mbs") ##prefix="%s-%s-%s" % (native, host, s))
                            result= measure(index, cmd, 0, 1000000, outFile=throwaway, errFile=output)
                            os.close(output)
                            if result.isOkay():
                                results.append( (native, host, s, path, result) )
                                index+= 1
                                retry= False
                            else:
                                print "(faulted; retrying)",

                    print
        print "Gathering data complete..."
        results= expand_file_information(results)

        print "Saving results to file ..."
        with open("res.out", 'w') as output:
            output.write(pickle.dumps(results))
    except:
        if os.path.exists("res.out"):
            print "Loading results from file ..."
            with open("res.out", 'r') as input:
                results= pickle.loads(input.read())

    tree= build_tree(results)

        # for item in tree.keys():
        #print "tree[ ", item, " ] = ", tree[ item ]
        #print item
#print "-------------------------------------------------------------------------------------------------------------------------------"

    verify_bmark_outputs(tree)
#    print_results_by_time(tree)
#    chart_results_by_time(tree)

    # test

    print_raw(tree)
    compacted_results = compact_results(tree)

#    order= [ ident for (ident, cmd_line, ___) in host_vms ]
    print_compacted_results(compacted_results)
    write_compacted_result_to_csv(compacted_results)

    ## 20110907/1128/sbr: TODO: should actually store the outputs using get_output_filename as a text representation...

    remove_temp_files(results)
    return []

def execute_script(script_filename, script_path=os.path.abspath(get_maxine_scripts())):
    """This function returns another closure so that we can bind the name of the script
    to be executed ("script_filename") to the execution closure. This is necessary because
    this filename comes from the OptionParser and needs to be made available to the
    execution closure in one way or another.
    """
    def closure():
        path= "%s/%s" % (script_path, script_filename)
        if not os.path.exists(path):
            exit('Specified script does not exist: %s' % path)

        yield get_exec(action_args=[path])
    return closure

def show_errors(errors, total_execs):
    print "#" * 80
    print "Summary:"
    print "\t%d out of %d execs failed:" % (len(errors), total_execs)
    for (executable, host_vm_opts, client_vm_opts, cmd_class, arguments) in errors:
        print "\t\t%s" % (arguments)

if __name__ == "__main__":
    check_config()
    (opts, args)= parse_cmd_line_args()

    driver= {
        ## binds a command line argument to an action...
        'build'   : do_build,
        'compile' : compile_gn,
        'execute' : execute_script(opts.execute),
        'run'     : run_gn,
        'bench'   : run_bmarks_gn,
        }

    ## the protocol is fairly easy and enables flexible arrangement of various actions
    ## and keeping passing of options and arguments hassle-free, too.
    ## the procedure is:
    ## 1) iterate over the driver-table and find which actions are specified,
    ## 2) execute the function specified in the table, these functions *MUST*
    ##    return an iterable data structure of callables (i.e., either a list or
    ##    a generator of functions),
    ## 3) call all functions of the list of step (2), supplying the action-independent
    ##    options and arguments.
    errors= []
    total_execs= 0
    for (key, iter_fn) in driver.iteritems():
        if hasattr(opts, key) and getattr(opts, key):
            for f in iter_fn():
                f(args=args, host_vm_opts=opts.host, client_vm_opts=opts.client,
                    use_maxine=opts.max, use_inspector=opts.ins, errors=errors)
                total_execs+= 1

    if errors:
        show_errors(errors, total_execs)


